"""The test runner"""

import os
import sys
from contextlib import contextmanager

from prysk._test import testfile

__all__ = ["runtests"]

if sys.platform == "win32":

    def _walk(top):
        top = os.fsdecode(top)
        for root, dirs, files in os.walk(top):
            yield (
                os.fsencode(root),
                [os.fsencode(p) for p in dirs],
                [os.fsencode(p) for p in files],
            )

else:
    _walk = os.walk


def _findtests(paths):
    """Yield tests in paths in sorted order"""
    for p in paths:
        if os.path.isdir(p):
            for root, _, files in _walk(p):
                if os.path.basename(root).startswith(b"."):
                    continue
                for f in sorted(files):
                    if not f.startswith(b".") and f.endswith(b".t"):
                        yield os.path.normpath(os.path.join(root, f))
        else:
            yield os.path.normpath(p)


@contextmanager
def _cwd(path):
    """Change the working directory

    This context manager will change the working directory
    and restore it afterwards.
    """
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(cwd)


def runtests(paths, tmpdir, shell, indent=2, cleanenv=True, debug=False):
    """Run tests and yield results.

    This yields a sequence of 2-tuples containing the following:

        (test path, test function)

    The test function, when called, runs the test in a temporary directory
    and returns a 3-tuple:

        (list of lines in the test, same list with actual output, diff)
    """
    seen = set()
    basenames = set()
    for i, path in enumerate(_findtests(paths)):
        abspath = os.path.abspath(path)
        if abspath in seen:
            continue
        seen.add(abspath)

        if not os.stat(path).st_size:
            yield path, lambda: (None, None, None)
            continue

        basename = os.path.basename(path)
        if basename in basenames:
            basename = basename + b"-%d" % i
        else:
            basenames.add(basename)

        def test():
            """Run test file"""
            testdir = os.path.join(tmpdir, basename)
            os.mkdir(testdir)
            with _cwd(testdir):
                return testfile(
                    abspath,
                    shell,
                    indent=indent,
                    cleanenv=cleanenv,
                    debug=debug,
                    testname=path,
                )

        yield path, test



"""xUnit XML output"""

import locale
import os
import re
import socket
import sys
import time

__all__ = ["runxunit"]

_widecdataregex = (
    r"(?:[^\x09\x0a\x0d\x20-\ud7ff\ue000-\ufffd" r"\U00010000-\U0010ffff]|]]>)"
)
_narrowcdataregex = r"(?:[^\x09\x0a\x0d\x20-\ud7ff\ue000-\ufffd]" r"|]]>)"
_widequoteattrregex = (
    r"[^\x20\x21\x23-\x25\x27-\x3b\x3d"
    r"\x3f-\ud7ff\ue000-\ufffd"
    r"\U00010000-\U0010ffff]"
)
_narrowquoteattrregex = r"[^\x20\x21\x23-\x25\x27-\x3b\x3d" r"\x3f-\ud7ff\ue000-\ufffd]"
_replacementchar = "\N{REPLACEMENT CHARACTER}"

if sys.maxunicode >= 0x10FFFF:
    _cdatasub = re.compile(_widecdataregex).sub
    _quoteattrsub = re.compile(_widequoteattrregex).sub
else:
    _cdatasub = re.compile(_narrowcdataregex).sub
    _quoteattrsub = re.compile(_narrowquoteattrregex).sub


def _cdatareplace(m):
    """Replace _cdatasub() regex match"""
    if m.group(0) == "]]>":
        return "]]>]]&gt;<![CDATA["
    else:
        return _replacementchar


def _cdata(s):
    r"""Escape a string as an XML CDATA block.

    >>> (_cdata('1<\'2\'>&"3\x00]]>\t\r\n') ==
    ...  '<![CDATA[1<\'2\'>&\"3\ufffd]]>]]&gt;<![CDATA[\t\r\n]]>')
    True
    """
    return "<![CDATA[%s]]>" % _cdatasub(_cdatareplace, s)


def _quoteattrreplace(m):
    """Replace _quoteattrsub() regex match"""
    return {
        "\t": "&#9;",
        "\n": "&#10;",
        "\r": "&#13;",
        '"': "&quot;",
        "&": "&amp;",
        "<": "&lt;",
        ">": "&gt;",
    }.get(m.group(0), _replacementchar)


def _quoteattr(s):
    r"""Escape a string for use as an XML attribute value.

    >>> (_quoteattr('1<\'2\'>&"3\x00]]>\t\r\n') ==
    ...  '"1&lt;\'2\'&gt;&amp;&quot;3\ufffd]]&gt;&#9;&#13;&#10;"')
    True
    """
    return '"%s"' % _quoteattrsub(_quoteattrreplace, s)


def _timestamp():
    """Return the current time in ISO 8601 format"""
    tm = time.localtime()
    if tm.tm_isdst == 1:
        tz = time.altzone
    else:
        tz = time.timezone

    timestamp = time.strftime("%Y-%m-%dT%H:%M:%S", tm)
    tzhours = int(-tz / 60 / 60)
    tzmins = int(abs(tz) / 60 % 60)
    timestamp += "%+03d:%02d" % (tzhours, tzmins)
    return timestamp


def runxunit(tests, xmlpath):
    """Run tests with xUnit XML output.

    tests should be a sequence of 2-tuples containing the following:

        (test path, test function)

    This function yields a new sequence where each test function is wrapped
    with a function that writes test results to an xUnit XML file.
    """
    suitestart = time.time()
    timestamp = _timestamp()
    hostname = socket.gethostname()
    total, skipped, failed = [0], [0], [0]
    testcases = []

    for path, test in tests:

        def testwrapper():
            """Run test and collect XML output"""
            total[0] += 1

            start = time.time()
            refout, postout, diff = test()
            testtime = time.time() - start

            classname = path.decode(locale.getpreferredencoding(), "replace")
            name = os.path.basename(classname)

            if postout is None:
                skipped[0] += 1
                testcase = (
                    "  <testcase classname=%(classname)s\n"
                    "            name=%(name)s\n"
                    '            time="%(time).6f">\n'
                    "    <skipped/>\n"
                    "  </testcase>\n"
                ) % {
                    "classname": _quoteattr(classname),
                    "name": _quoteattr(name),
                    "time": testtime,
                }
            elif diff:
                failed[0] += 1
                diff = list(diff)
                diffu = "".join(
                    l.decode(locale.getpreferredencoding(), "replace") for l in diff
                )
                testcase = (
                    "  <testcase classname=%(classname)s\n"
                    "            name=%(name)s\n"
                    '            time="%(time).6f">\n'
                    "    <failure>%(diff)s</failure>\n"
                    "  </testcase>\n"
                ) % {
                    "classname": _quoteattr(classname),
                    "name": _quoteattr(name),
                    "time": testtime,
                    "diff": _cdata(diffu),
                }
            else:
                testcase = (
                    "  <testcase classname=%(classname)s\n"
                    "            name=%(name)s\n"
                    '            time="%(time).6f"/>\n'
                ) % {
                    "classname": _quoteattr(classname),
                    "name": _quoteattr(name),
                    "time": testtime,
                }
            testcases.append(testcase)

            return refout, postout, diff

        yield path, testwrapper

    suitetime = time.time() - suitestart
    header = (
        '<?xml version="1.0" encoding="utf-8"?>\n'
        '<testsuite name="prysk"\n'
        '           tests="%(total)d"\n'
        '           failures="%(failed)d"\n'
        '           skipped="%(skipped)d"\n'
        "           timestamp=%(timestamp)s\n"
        "           hostname=%(hostname)s\n"
        '           time="%(time).6f">\n'
    ) % {
        "total": total[0],
        "failed": failed[0],
        "skipped": skipped[0],
        "timestamp": _quoteattr(timestamp),
        "hostname": _quoteattr(hostname),
        "time": suitetime,
    }
    footer = "</testsuite>\n"

    xmlfile = open(xmlpath, "wb")
    try:
        xmlfile.write(header.encode("utf-8"))
        for testcase in testcases:
            xmlfile.write(testcase.encode("utf-8"))
        xmlfile.write(footer.encode("utf-8"))
    finally:
        xmlfile.close()



"""Main entry point"""

import configparser
import optparse
import os
import shlex
import shutil
import sys
import tempfile

from prysk._run import runtests
from prysk._xunit import runxunit
from prysk.cli import runcli


def _which(cmd):
    """Return the path to cmd or None if not found"""
    cmd = os.fsencode(cmd)
    for p in os.environ["PATH"].split(os.pathsep):
        path = os.path.join(os.fsencode(p), cmd)
        if os.path.isfile(path) and os.access(path, os.X_OK):
            return os.path.abspath(path)
    return None


def _expandpath(path):
    """Expands ~ and environment variables in path"""
    return os.path.expanduser(os.path.expandvars(path))


class _OptionParser(optparse.OptionParser):
    """Like optparse.OptionParser, but supports setting values through
    PRYSK= and .pryskrc."""

    def __init__(self, *args, **kwargs):
        self._config_opts = {}
        optparse.OptionParser.__init__(self, *args, **kwargs)

    def add_option(self, *args, **kwargs):
        option = optparse.OptionParser.add_option(self, *args, **kwargs)
        if option.dest and option.dest != "version":
            key = option.dest.replace("_", "-")
            self._config_opts[key] = option.action == "store_true"
        return option

    def parse_args(self, args=None, values=None):
        config = configparser.RawConfigParser()
        config.read(_expandpath(os.environ.get("PRYSKRC", ".pryskrc")))
        defaults = {}
        for key, isbool in self._config_opts.items():
            try:
                if isbool:
                    try:
                        value = config.getboolean("prysk", key)
                    except ValueError:
                        value = config.get("prysk", key)
                        self.error("--%s: invalid boolean value: %r" % (key, value))
                else:
                    value = config.get("prysk", key)
            except (configparser.NoSectionError, configparser.NoOptionError):
                pass
            else:
                defaults[key] = value
        self.set_defaults(**defaults)

        eargs = os.environ.get("PRYSK", "").strip()
        if eargs:
            args = args or []
            args += shlex.split(eargs)

        try:
            return optparse.OptionParser.parse_args(self, args, values)
        except optparse.OptionValueError:
            self.error(str(sys.exc_info()[1]))


def _parseopts(args):
    """Parse command line arguments"""
    p = _OptionParser(usage="prysk [OPTIONS] TESTS...", prog="prysk")
    p.add_option(
        "-V", "--version", action="store_true", help="show version information and exit"
    )
    p.add_option("-q", "--quiet", action="store_true", help="don't print diffs")
    p.add_option(
        "-v", "--verbose", action="store_true", help="show filenames and test status"
    )
    p.add_option(
        "-i",
        "--interactive",
        action="store_true",
        help="interactively merge changed test output",
    )
    p.add_option(
        "-d",
        "--debug",
        action="store_true",
        help="write script output directly to the terminal",
    )
    p.add_option("-y", "--yes", action="store_true", help="answer yes to all questions")
    p.add_option("-n", "--no", action="store_true", help="answer no to all questions")
    p.add_option(
        "-E",
        "--preserve-env",
        action="store_true",
        help="don't reset common environment variables",
    )
    p.add_option(
        "--keep-tmpdir", action="store_true", help="keep temporary directories"
    )
    p.add_option(
        "--shell",
        action="store",
        default="/bin/sh",
        metavar="PATH",
        help="shell to use for running tests (default: %default)",
    )
    p.add_option(
        "--shell-opts",
        action="store",
        metavar="OPTS",
        help="arguments to invoke shell with",
    )
    p.add_option(
        "--indent",
        action="store",
        default=2,
        metavar="NUM",
        type="int",
        help=("number of spaces to use for indentation " "(default: %default)"),
    )
    p.add_option(
        "--xunit-file",
        action="store",
        metavar="PATH",
        help="path to write xUnit XML output",
    )
    opts, paths = p.parse_args(args)
    paths = [os.fsencode(path) for path in paths]
    return opts, paths, p.get_usage


def main(args):
    """Main entry point.

    If you're thinking of using Cram in other Python code (e.g., unit tests),
    consider using the test() or testfile() functions instead.

    :param args: Script arguments (excluding script name)
    :type args: str
    :return: Exit code (non-zero on failure)
    :rtype: int
    """
    opts, paths, getusage = _parseopts(args)
    if opts.version:
        sys.stdout.write(
            """Cram CLI testing framework (version 0.8)

Copyright (C) 2010-2021 Brodie Rao <brodie@bitheap.org> and others
This is free software; see the source for copying conditions. There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
"""
        )
        return

    conflicts = [
        ("--yes", opts.yes, "--no", opts.no),
        ("--quiet", opts.quiet, "--interactive", opts.interactive),
        ("--debug", opts.debug, "--quiet", opts.quiet),
        ("--debug", opts.debug, "--interactive", opts.interactive),
        ("--debug", opts.debug, "--verbose", opts.verbose),
        ("--debug", opts.debug, "--xunit-file", opts.xunit_file),
    ]
    for s1, o1, s2, o2 in conflicts:
        if o1 and o2:
            sys.stderr.write(f"options {s1} and {s2} are mutually exclusive\n")
            return 2

    shellcmd = _which(opts.shell)
    if not shellcmd:
        sys.stderr.buffer.write(b"shell not found: %s\n" % os.fsencode(opts.shell))
        return 2
    shell = [shellcmd]
    if opts.shell_opts:
        shell += shlex.split(opts.shell_opts)

    patchcmd = None
    if opts.interactive:
        patchcmd = _which("patch")
        if not patchcmd:
            sys.stderr.write("patch(1) required for -i\n")
            return 2

    if not paths:
        sys.stdout.write(getusage())
        return 2

    badpaths = [path for path in paths if not os.path.exists(path)]
    if badpaths:
        sys.stderr.buffer.write(b"no such file: %s\n" % badpaths[0])
        return 2

    if opts.yes:
        answer = "y"
    elif opts.no:
        answer = "n"
    else:
        answer = None

    tmpdir = os.environ["PRYSK_TEMP"] = tempfile.mkdtemp("", "prysk-tests-")
    tmpdirb = os.fsencode(tmpdir)
    proctmp = os.path.join(tmpdir, "tmp")
    for s in ("TMPDIR", "TEMP", "TMP"):
        os.environ[s] = proctmp

    os.mkdir(proctmp)
    try:
        tests = runtests(
            paths,
            tmpdirb,
            shell,
            indent=opts.indent,
            cleanenv=not opts.preserve_env,
            debug=opts.debug,
        )
        if not opts.debug:
            tests = runcli(
                tests,
                quiet=opts.quiet,
                verbose=opts.verbose,
                patchcmd=patchcmd,
                answer=answer,
            )
            if opts.xunit_file is not None:
                tests = runxunit(tests, opts.xunit_file)

        hastests = False
        failed = False
        for path, test in tests:
            hastests = True
            _, _, diff = test()
            if diff:
                failed = True

        if not hastests:
            sys.stderr.write("no tests found\n")
            return 2

        return int(failed)
    finally:
        if opts.keep_tmpdir:
            sys.stdout.buffer.write(b"# Kept temporary directory: %s\n" % tmpdirb)
        else:
            shutil.rmtree(tmpdir)



"""The command line interface implementation"""
import os
import sys

import prysk
from prysk._process import execute

__all__ = ["runcli"]


def _prompt(question, answers, auto=None):
    """Write a prompt to stdout and ask for answer in stdin.

    answers should be a string, with each character a single
    answer. An uppercase letter is considered the default answer.

    If an invalid answer is given, this asks again until it gets a
    valid one.

    If auto is set, the question is answered automatically with the
    specified value.
    """
    default = [c for c in answers if c.isupper()]
    while True:
        sys.stdout.write(f"{question} [{answers}] ")
        sys.stdout.flush()
        if auto is not None:
            sys.stdout.write(auto + "\n")
            sys.stdout.flush()
            return auto

        answer = sys.stdin.readline().strip().lower()
        if not answer and default:
            return default[0]
        elif answer and answer in answers.lower():
            return answer


def _log(msg=None, verbosemsg=None, verbose=False):
    """Write msg to standard out and flush.

    If verbose is True, write verbosemsg instead.
    """
    if verbose:
        msg = verbosemsg
    if msg:
        if isinstance(msg, bytes):
            sys.stdout.buffer.write(msg)
        else:
            sys.stdout.write(msg)
        sys.stdout.flush()


def _patch(cmd, diff):
    """Run echo [lines from diff] | cmd -p0"""
    _, retcode = execute([cmd, "-p0"], stdin=b"".join(diff))
    return retcode == 0


def runcli(tests, quiet=False, verbose=False, patchcmd=None, answer=None):
    """Run tests with command line interface input/output.

    tests should be a sequence of 2-tuples containing the following:

        (test path, test function)

    This function yields a new sequence where each test function is wrapped
    with a function that handles CLI input/output.

    If quiet is True, diffs aren't printed. If verbose is True,
    filenames and status information are printed.

    If patchcmd is set, a prompt is written to stdout asking if
    changed output should be merged back into the original test. The
    answer is read from stdin. If 'y', the test is patched using patch
    based on the changed output.
    """
    total, skipped, failed = [0], [0], [0]

    for path, test in tests:

        def testwrapper():
            """Test function that adds CLI output"""
            total[0] += 1
            _log(None, path + b": ", verbose)

            refout, postout, diff = test()
            if refout is None:
                skipped[0] += 1
                _log("s", "empty\n", verbose)
                return refout, postout, diff

            abspath = os.path.abspath(path)
            errpath = abspath + b".err"

            if postout is None:
                skipped[0] += 1
                _log("s", "skipped\n", verbose)
            elif not diff:
                _log(".", "passed\n", verbose)
                if os.path.exists(errpath):
                    os.remove(errpath)
            else:
                failed[0] += 1
                _log("!", "failed\n", verbose)
                if not quiet:
                    _log("\n", None, verbose)

                with open(errpath, "wb") as errfile:
                    for line in postout:
                        errfile.write(line)

                if not quiet:
                    origdiff = diff
                    diff = []
                    for line in origdiff:
                        sys.stdout.buffer.write(line)
                        diff.append(line)

                    if patchcmd and _prompt("Accept this change?", "yN", answer) == "y":
                        if _patch(patchcmd, diff):
                            _log(None, path + b": merged output\n", verbose)
                            os.remove(errpath)
                        else:
                            _log(path + b": merge failed\n")

            return refout, postout, diff

        yield path, testwrapper

    if total[0] > 0:
        _log("\n", None, verbose)
        _log(f"# Ran {total[0]} tests, {skipped[0]} skipped, {failed[0]} failed.\n")


def main():
    try:
        sys.exit(prysk.main(sys.argv[1:]))
    except (BrokenPipeError, KeyboardInterrupt):
        pass


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prysk']

package_data = \
{'': ['*']}

install_requires = \
['pytest>=7.0.1,<8.0.0', 'tomli-w>=1.0.0,<2.0.0', 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['prysk = prysk:cli.main']}

setup_kwargs = {
    'name': 'prysk',
    'version': '0.10.0',
    'description': 'Functional tests for command line applications',
    'long_description': 'Prysk\n======================\n.. image:: https://github.com/Nicoretti/prysk/actions/workflows/verifier.yaml/badge.svg\n    :target: https://github.com/Nicoretti/prysk/actions/workflows/verifier.yaml\n\n.. image:: https://coveralls.io/repos/github/Nicoretti/prysk/badge.svg?branch=master\n    :target: https://coveralls.io/github/Nicoretti/prysk?branch=master\n\n.. image:: https://img.shields.io/badge/imports-isort-ef8336.svg\n    :target: https://pycqa.github.io/isort/\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n\n.. image:: https://img.shields.io/badge/pypi%20package-available-blue.svg\n    :target: https://pypi.org/project/prysk/\n\n.. image:: https://img.shields.io/badge/docs-available-blue.svg\n    :target: https://nicoretti.github.io/prysk/\n\nPrysk is a fork of the popular snapshot testing tool Cram_.\nEven though Cram_ is pretty complete and mature for everyday use,\nPrysk want\'s to continue pushing it\'s development forward.\n\n.. _Cram: https://bitheap.org/cram\n\nPrysk tests look like snippets of interactive shell sessions. Prysk runs\neach command and compares the command output in the test with the\ncommand\'s actual output.\n\nHere\'s a snippet from `Prysk\'s own test suite`_::\n\n    Set up prysk alias and example tests:\n\n      $ . "$TESTDIR"/setup.sh\n\n    Usage:\n\n      $ prysk -h\n      [Uu]sage: prysk \\[OPTIONS\\] TESTS\\.\\.\\. (re)\n\n      [Oo]ptions: (re)\n        -h, --help          show this help message and exit\n        -V, --version       show version information and exit\n        -q, --quiet         don\'t print diffs\n        -v, --verbose       show filenames and test status\n        -i, --interactive   interactively merge changed test output\n        -d, --debug         write script output directly to the terminal\n        -y, --yes           answer yes to all questions\n        -n, --no            answer no to all questions\n        -E, --preserve-env  don\'t reset common environment variables\n        --keep-tmpdir       keep temporary directories\n        --shell=PATH        shell to use for running tests (default: /bin/sh)\n        --shell-opts=OPTS   arguments to invoke shell with\n        --indent=NUM        number of spaces to use for indentation (default: 2)\n        --xunit-file=PATH   path to write xUnit XML output\n\nThe format in a nutshell:\n\n* Prysk tests use the ``.t`` file extension.\n\n* Lines beginning with two spaces, a dollar sign, and a space are run\n  in the shell.\n\n* Lines beginning with two spaces, a greater than sign, and a space\n  allow multi-line commands.\n\n* All other lines beginning with two spaces are considered command\n  output.\n\n* Output lines ending with a space and the keyword ``(re)`` are\n  matched as `Perl-compatible regular expressions`_.\n\n* Lines ending with a space and the keyword ``(glob)`` are matched\n  with a glob-like syntax. The only special characters supported are\n  ``*`` and ``?``. Both characters can be escaped using ``\\``, and the\n  backslash can be escaped itself.\n\n* Output lines ending with either of the above keywords are always\n  first matched literally with actual command output.\n\n* Lines ending with a space and the keyword ``(no-eol)`` will match\n  actual output that doesn\'t end in a newline.\n\n* Actual output lines containing unprintable characters are escaped\n  and suffixed with a space and the keyword ``(esc)``. Lines matching\n  unprintable output must also contain the keyword.\n\n* Anything else is a comment.\n\n.. _Prysk\'s own test suite: https://github.com/nicoretti/prysk/blob/master/tests/usage.t\n.. _Perl-compatible regular expressions: https://en.wikipedia.org/wiki/Perl_Compatible_Regular_Expressions\n\nUsage\n-----\n\nPrysk will print a dot for each passing test. If a test fails, a\n`unified context diff`_ is printed showing the test\'s expected output\nand the actual output. Skipped tests (empty tests and tests that exit\nwith return code ``80``) are marked with ``s`` instead of a dot.\n\nFor example, if we run Prysk on `its own example tests`_::\n\n    .s.!\n    --- examples/fail.t\n    +++ examples/fail.t.err\n    @@ -3,21 +3,22 @@\n       $ echo 1\n       1\n       $ echo 1\n    -  2\n    +  1\n       $ echo 1\n       1\n\n     Invalid regex:\n\n       $ echo 1\n    -  +++ (re)\n    +  1\n\n     Offset regular expression:\n\n       $ printf \'foo\\nbar\\nbaz\\n\\n1\\nA\\n@\\n\'\n       foo\n    +  bar\n       baz\n\n       \\d (re)\n       [A-Z] (re)\n    -  #\n    +  @\n    s.\n    # Ran 6 tests, 2 skipped, 1 failed.\n\nPrysk will also write the test with its actual output to\n``examples/fail.t.err``, allowing you to use other diff tools. This\nfile is automatically removed the next time the test passes.\n\nWhen you\'re first writing a test, you might just write the commands\nand run the test to see what happens. If you run Prysk with ``-i`` or\n``--interactive``, you\'ll be prompted to merge the actual output back\ninto the test. This makes it easy to quickly prototype new tests.\n\nYou can specify a default set of options by creating a ``.prysk``\nfile. For example::\n\n    [prysk]\n    verbose = True\n    indent = 4\n\nIs the same as invoking Prysk with ``--verbose`` and ``--indent=4``.\n\nTo change what configuration file Prysk loads, you can set the\n``PRYSKRC`` environment variable. You can also specify command line\noptions in the ``PRYSK`` environment variable.\n\nNote that the following environment variables are reset before tests\nare run:\n\n* ``TMPDIR``, ``TEMP``, and ``TMP`` are set to the test runner\'s\n  ``tmp`` directory.\n\n* ``LANG``, ``LC_ALL``, and ``LANGUAGE`` are set to ``C``.\n\n* ``TZ`` is set to ``GMT``.\n\n* ``COLUMNS`` is set to ``80``. (Note: When using ``--shell=zsh``,\n  this cannot be reset. It will reflect the actual terminal\'s width.)\n\n* ``CDPATH`` and ``GREP_OPTIONS`` are set to an empty string.\n\nPrysk also provides the following environment variables to tests:\n\n* ``PRYSK_TEMP``, set to the test runner\'s temporary directory.\n\n* ``TESTDIR``, set to the directory containing the test file.\n\n* ``TESTFILE``, set to the basename of the current test file.\n\n* ``TESTSHELL``, set to the value specified by ``--shell``.\n\nAlso note that care should be taken with commands that close the test\nshell\'s ``stdin``. For example, if you\'re trying to invoke ``ssh`` in\na test, try adding the ``-n`` option to prevent it from closing\n``stdin``. Similarly, if you invoke a daemon process that inherits\n``stdout`` and fails to close it, it may cause Prysk to hang while\nwaiting for the test shell\'s ``stdout`` to be fully closed.\n\n.. _unified context diff: https://en.wikipedia.org/wiki/Diff#Unified_format\n.. _its own example tests: https://github.com/nicoretti/prysk/tree/master/examples\n',
    'author': 'Nicola Coretti',
    'author_email': 'nico.coretti@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://nicoretti.github.io/prysk/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Utilities for diffing test files and their output"""

import codecs
import difflib
import re

__all__ = ["esc", "glob", "regex", "unified_diff"]


def _regex(pattern, s):
    """Match a regular expression or return False if invalid.

    >>> [bool(_regex(r, b'foobar')) for r in (b'foo.*', b'***')]
    [True, False]
    """
    try:
        return re.match(pattern + rb"\Z", s)
    except re.error:
        return False


def _glob(el, l):
    r"""Match a glob-like pattern.

    The only supported special characters are * and ?. Escaping is
    supported.

    >>> bool(_glob(br'\* \\ \? fo?b*', b'* \\ ? foobar'))
    True
    """
    i, n = 0, len(el)
    res = b""
    while i < n:
        c = el[i : i + 1]
        i += 1
        if c == b"\\" and el[i] in b"*?\\":
            res += el[i - 1 : i + 1]
            i += 1
        elif c == b"*":
            res += b".*"
        elif c == b"?":
            res += b"."
        else:
            res += re.escape(c)
    return _regex(res, l)


def _matchannotation(keyword, matchfunc, el, l):
    """Apply match function based on annotation keyword"""
    ann = b" (%s)\n" % keyword
    return el.endswith(ann) and matchfunc(el[: -len(ann)], l[:-1])


def regex(el, l):
    """Apply a regular expression match to a line annotated with '(re)'"""
    return _matchannotation(b"re", _regex, el, l)


def glob(el, l):
    """Apply a glob match to a line annotated with '(glob)'"""
    return _matchannotation(b"glob", _glob, el, l)


def esc(el, l):
    """Apply an escape match to a line annotated with '(esc)'"""
    ann = b" (esc)\n"

    if el.endswith(ann):
        el = codecs.escape_decode(el[: -len(ann)])[0] + b"\n"
    if el == l:
        return True

    if l.endswith(ann):
        l = codecs.escape_decode(l[: -len(ann)])[0] + b"\n"
    return el == l


class _SequenceMatcher(difflib.SequenceMatcher):
    """Like difflib.SequenceMatcher, but supports custom match functions"""

    def __init__(self, *args, **kwargs):
        self._matchers = kwargs.pop("matchers", [])
        super().__init__(*args, **kwargs)

    def _match(self, el, l):
        """Tests for matching lines using custom matchers"""
        for matcher in self._matchers:
            if matcher(el, l):
                return True
        return False

    def find_longest_match(self, alo, ahi, blo, bhi):
        """Find longest matching block in a[alo:ahi] and b[blo:bhi]"""
        # SequenceMatcher uses find_longest_match() to slowly whittle down
        # the differences between a and b until it has each matching block.
        # Because of this, we can end up doing the same matches many times.
        matches = []
        for n, (el, line) in enumerate(zip(self.a[alo:ahi], self.b[blo:bhi])):
            if el != line and self._match(el, line):
                # This fools the superclass's method into thinking that the
                # regex/glob in a is identical to b by replacing a's line (the
                # expected output) with b's line (the actual output).
                self.a[alo + n] = line
                matches.append((n, el))
        ret = super().find_longest_match(alo, ahi, blo, bhi)
        # Restore the lines replaced above. Otherwise, the diff output
        # would seem to imply that the tests never had any regexes/globs.
        for n, el in matches:
            self.a[alo + n] = el
        return ret


def unified_diff(
    l1,
    l2,
    fromfile=b"",
    tofile=b"",
    fromfiledate=b"",
    tofiledate=b"",
    n=3,
    lineterm=b"\n",
    matchers=None,
):
    r"""Compare two sequences of lines; generate the delta as a unified diff.

    This is like difflib.unified_diff(), but allows custom matchers.

    >>> l1 = [b'a\n', b'? (glob)\n']
    >>> l2 = [b'a\n', b'b\n']
    >>> (list(unified_diff(l1, l2, b'f1', b'f2', b'1970-01-01',
    ...                    b'1970-01-02')) ==
    ...  [b'--- f1\t1970-01-01\n', b'+++ f2\t1970-01-02\n',
    ...   b'@@ -1,2 +1,2 @@\n', b' a\n', b'-? (glob)\n', b'+b\n'])
    True

    >>> from prysk._diff import glob
    >>> list(unified_diff(l1, l2, matchers=[glob]))
    []
    """
    if matchers is None:
        matchers = []
    started = False
    matcher = _SequenceMatcher(None, l1, l2, matchers=matchers)
    for group in matcher.get_grouped_opcodes(n):
        if not started:
            if fromfiledate:
                fromdate = b"\t" + fromfiledate
            else:
                fromdate = b""
            if tofiledate:
                todate = b"\t" + tofiledate
            else:
                todate = b""
            yield b"--- " + fromfile + fromdate + lineterm
            yield b"+++ " + tofile + todate + lineterm
            started = True
        i1, i2, j1, j2 = group[0][1], group[-1][2], group[0][3], group[-1][4]
        yield b"@@ -%d,%d +%d,%d @@" % (i1 + 1, i2 - i1, j1 + 1, j2 - j1) + lineterm
        for tag, i1, i2, j1, j2 in group:
            if tag == "equal":
                for line in l1[i1:i2]:
                    yield b" " + line
                continue
            if tag in ("replace", "delete"):
                for line in l1[i1:i2]:
                    yield b"-" + line
            if tag in ("replace", "insert"):
                for line in l2[j1:j2]:
                    yield b"+" + line



from pathlib import Path

import nox

BASEPATH = Path(__file__).parent.resolve()

nox.options.sessions = [
    "isort",
    "code_format",
    "pylint",
    "unit",
    "integration",
    "coverage",
    "docs",
]


@nox.session(python=False)
def code_format(session):
    session.run(
        "poetry",
        "run",
        "python",
        "-m",
        "black",
        "--check",
        "--diff",
        "--color",
        f"{BASEPATH}",
    )


@nox.session(python=False)
def isort(session):
    session.run(
        "poetry", "run", "python", "-m", "isort", "-v", "--check", f"{BASEPATH}"
    )


@nox.session(python=False)
def pylint(session):
    session.run("poetry", "run", "python", "-m", "pylint", f'{BASEPATH / "prysk"}')
    session.run("poetry", "run", "python", "-m", "pylint", f'{BASEPATH / "scripts"}')


@nox.session(python=False)
def unit(session):
    session.run(
        "poetry",
        "run",
        "python",
        "-m",
        "pytest",
        "--ignore-glob=prysk/*.py",
        f"{BASEPATH}",
    )


@nox.session
@nox.parametrize("shell", ["dash", "bash", "zsh"])
def integration(session, shell):
    session.install(f"{BASEPATH}")
    session.env["TESTOPTS"] = f"--shell={shell}"
    session.run(
        "poetry",
        "run",
        "prysk",
        f"--shell={shell}",
        f'{BASEPATH / "test" / "integration" / "prysk"}',
    )


@nox.session(python=False)
def coverage(session):
    session.env["COVERAGE"] = "coverage"
    session.env["COVERAGE_FILE"] = f'{BASEPATH / ".coverage"}'
    command = [
        "poetry",
        "run",
        "coverage",
        "run",
        "-a",
        f'--rcfile={BASEPATH / "pyproject.toml"}',
        "-m",
        "prysk.cli",
    ]
    session.run(
        *(command + ["--shell=bash", f'{BASEPATH / "test" / "integration" / "prysk"}'])
    )
    session.run(
        *(command + ["--shell=dash", f'{BASEPATH / "test" / "integration" / "prysk"}'])
    )
    session.run(
        *(command + ["--shell=zsh", f'{BASEPATH / "test" / "integration" / "prysk"}'])
    )
    session.run("coverage", "report", "--fail-under=97")
    session.run("coverage", "lcov")


@nox.session(python=False)
def docs(session):
    session.run("make", "-C", f'{BASEPATH / "docs"}', "html")



""" Cli tool to calculate the md5 checksum for one or multiple files """
import argparse
import sys
from hashlib import md5


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "files", metavar="FILE", type=argparse.FileType("rb"), nargs="+"
    )
    return parser


def main(argv=None):
    parser = _create_parser()
    args = parser.parse_args(argv)
    for file in args.files:
        hasher = md5()
        hasher.update(file.read())
        print(hasher.hexdigest())


if __name__ == "__main__":
    sys.exit(main())
